@@ -1,6 +1,6 @@
 /*
     Scan and print all the PC/SC devices available
-    Copyright (C) 2001-2008  Ludovic Rousseau <ludovic.rousseau@free.fr>
+    Copyright (C) 2001-2009  Ludovic Rousseau <ludovic.rousseau@free.fr>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -17,7 +17,7 @@
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */
 
-/* $Id: pcsc_scan.c,v 1.30 2008-03-23 17:05:20 rousseau Exp $ */
+/* $Id: pcsc_scan.c,v 1.45 2009-01-06 21:06:48 rousseau Exp $ */
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -37,11 +37,30 @@
 #define FALSE 0
 #endif
 
+/* use "\\?PnP?\Notification" mechanism when possible */
+#ifndef __APPLE__
+#define PNP
+#endif
+
 #define TIMEOUT 1000	/* 1 second timeout */
 
 /* command used to parse (on screen) the ATR */
 #define ATR_PARSER "ATR_analysis"
 
+#ifndef SCARD_E_NO_READERS_AVAILABLE
+#define SCARD_E_NO_READERS_AVAILABLE 0x8010002E
+#endif
+
+#define test_rv(fct, rv, hContext) \
+do { \
+	if (rv != SCARD_S_SUCCESS) \
+	{ \
+		printf("%s%s: %s%s\n", red, fct, pcsc_stringify_error(rv), color_end); \
+		(void)SCardReleaseContext(hContext); \
+		exit(-1); \
+	} \
+} while(0)
+
 void usage(void)
 {
 	printf("usage: pcsc_scan [-n] [-V] [-h]\n");
@@ -56,7 +75,8 @@ int main(int argc, char *argv[])
 	LONG rv;
 	SCARDCONTEXT hContext;
 	SCARD_READERSTATE_A *rgReaderStates_t = NULL;
-	DWORD dwReaders, dwReadersOld;
+	DWORD dwReaders = 0, dwReadersOld;
+	DWORD timeout;
 	LPSTR mszReaders = NULL;
 	char *ptr, **readers = NULL;
 	int nbReaders, i;
@@ -70,7 +90,7 @@ int main(int argc, char *argv[])
 	char *color_end = "";
 
 	printf("PC/SC device scanner\n");
-	printf("V " VERSION " (c) 2001-2008, Ludovic Rousseau <ludovic.rousseau@free.fr>\n");
+	printf("V " VERSION " (c) 2001-2009, Ludovic Rousseau <ludovic.rousseau@free.fr>\n");
 	printf("Compiled with PC/SC lite version: " PCSCLITE_VERSION_NUMBER "\n");
 
 	while ((opt = getopt(argc, argv, "Vhn")) != EOF)
@@ -124,15 +144,10 @@ int main(int argc, char *argv[])
 				}
 			}
 		}
-
 	}
 
 	rv = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &hContext);
-	if (rv != SCARD_S_SUCCESS)
-	{
-		printf("SCardEstablishContext: Cannot Connect to Resource Manager: %s (0x%lX)\n", pcsc_stringify_error(rv), rv);
-		return 1;
-	}
+	test_rv("SCardEstablishContext", rv, hContext);
 
 get_readers:
 	/* free memory possibly allocated in a previous loop */
@@ -154,13 +169,11 @@ get_readers:
 	 * 2. malloc the necessary storage
 	 * 3. call with the real allocated buffer
 	 */
-	printf("%sScanning present readers%s\n", magenta, color_end);
+	printf("%sScanning present readers...%s\n", red, color_end);
 	rv = SCardListReaders(hContext, NULL, NULL, &dwReaders);
-	if ((rv != SCARD_S_SUCCESS) && (rv != SCARD_E_NO_READERS_AVAILABLE))
-	{
-		printf("SCardListReader: %s (0x%lX)\n", pcsc_stringify_error(rv), rv);
-		exit(-1);
-	}
+	if (rv != SCARD_E_NO_READERS_AVAILABLE)
+		test_rv("SCardListReaders", rv, hContext);
+
 	dwReadersOld = dwReaders;
 
 	/* if non NULL we came back so free first */
@@ -179,10 +192,6 @@ get_readers:
 
 	*mszReaders = '\0';
 	rv = SCardListReaders(hContext, NULL, mszReaders, &dwReaders);
-	if (rv != SCARD_S_SUCCESS)
-	{
-		printf("SCardListReader: %s (0x%lX)\n", pcsc_stringify_error(rv), rv);
-	}
 
 	/* Extract readers from the null separated string and get the total
 	 * number of readers */
@@ -194,10 +203,20 @@ get_readers:
 		nbReaders++;
 	}
 
-	if (nbReaders == 0)
+	if (SCARD_E_NO_READERS_AVAILABLE == rv || 0 == nbReaders)
 	{
+#ifdef PNP
+		SCARD_READERSTATE_A rgReaderStates[1];
+#endif
 		printf("%sWaiting for the first reader...%s", red, color_end);
 		fflush(stdout);
+#ifdef PNP
+		rgReaderStates[0].szReader = "\\\\?PnP?\\Notification";
+		rgReaderStates[0].dwCurrentState = SCARD_STATE_UNAWARE;
+
+		rv = SCardGetStatusChange(hContext, INFINITE, rgReaderStates, 1);
+		test_rv("SCardGetStatusChange", rv, hContext);
+#else
 		rv = SCARD_S_SUCCESS;
 		while ((SCARD_S_SUCCESS == rv) && (dwReaders == dwReadersOld))
 		{
@@ -206,12 +225,15 @@ get_readers:
 				rv = SCARD_S_SUCCESS;
 			sleep(1);
 		}
+#endif
 		printf("found one\n");
 		goto get_readers;
 	}
+	else
+		test_rv("SCardListReader", rv, hContext);
 
 	/* allocate the readers table */
-	readers = calloc(nbReaders, sizeof(char *));
+	readers = calloc(nbReaders+1, sizeof(char *));
 	if (NULL == readers)
 	{
 		printf("Not enough memory for readers table\n");
@@ -230,7 +252,7 @@ get_readers:
 	}
 
 	/* allocate the ReaderStates table */
-	rgReaderStates_t = calloc(nbReaders, sizeof(* rgReaderStates_t));
+	rgReaderStates_t = calloc(nbReaders+1, sizeof(* rgReaderStates_t));
 	if (NULL == rgReaderStates_t)
 	{
 		printf("Not enough memory for readers states\n");
@@ -245,17 +267,31 @@ get_readers:
 		rgReaderStates_t[i].szReader = readers[i];
 		rgReaderStates_t[i].dwCurrentState = SCARD_STATE_UNAWARE;
 	}
+	rgReaderStates_t[nbReaders].szReader = "\\\\?PnP?\\Notification";
+	rgReaderStates_t[nbReaders].dwCurrentState = SCARD_STATE_UNAWARE;
 
 	/* Wait endlessly for all events in the list of readers
 	 * We only stop in case of an error
 	 */
-	rv = SCardGetStatusChange(hContext, TIMEOUT, rgReaderStates_t, nbReaders);
+#ifdef PNP
+	timeout = INFINITE;
+	nbReaders++;
+#else
+	timeout = TIMEOUT;
+#endif
+	rv = SCardGetStatusChange(hContext, timeout, rgReaderStates_t, nbReaders);
 	while ((rv == SCARD_S_SUCCESS) || (rv == SCARD_E_TIMEOUT))
 	{
+#ifdef PNP
+		if (rgReaderStates_t[nbReaders-1].dwEventState &
+			SCARD_STATE_CHANGED)
+			goto get_readers;
+#else
 		/* A new reader appeared? */
 		if ((SCardListReaders(hContext, NULL, NULL, &dwReaders)
 			== SCARD_S_SUCCESS) && (dwReaders != dwReadersOld))
 				goto get_readers;
+#endif
 
 		/* Now we have an event, check all the readers in the list to see what
 		 * happened */
@@ -366,18 +402,16 @@ get_readers:
 
 		} /* for */
 
-		rv = SCardGetStatusChange(hContext, TIMEOUT, rgReaderStates_t, nbReaders);
+		rv = SCardGetStatusChange(hContext, timeout, rgReaderStates_t,
+			nbReaders);
 	} /* while */
 
 	/* If we get out the loop, GetStatusChange() was unsuccessful */
-	printf("SCardGetStatusChange: %s (0x%lX)\n", pcsc_stringify_error(rv), rv);
+	test_rv("SCardGetStatusChange", rv, hContext);
 	
 	/* We try to leave things as clean as possible */
 	rv = SCardReleaseContext(hContext);
-	if (rv != SCARD_S_SUCCESS)
-	{
-		printf("SCardReleaseContext: %s (0x%lX)\n", pcsc_stringify_error(rv), rv);
-	}
+	test_rv("SCardReleaseContext", rv, hContext);
 
 	/* free memory possibly allocated */
 	if (NULL != readers)
